/*
 * gapi_app2d.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "gapi_app2d.hpp"

//
// static "self" ptr
//
GapiApp2D * GapiApp2D::self_ptr = 0;

//
// window procedure callback
//

LRESULT CALLBACK GapiApp2D_WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return GapiApp2D::self()->wndProc(hwnd, msg, wparam, lparam);
}


//
// ctor
//

template<class T> inline void MyImport(T& target, HMODULE lib, LPCWSTR name)
{
    target = (T)(GetProcAddressW(lib, name));
}

GapiApp2D::GapiApp2D(const TCHAR * window_class_name_, const TCHAR * window_title_)
    : window_class_name(window_class_name_), window_title(window_title_),
      screen_width(0), screen_height(0), display_open(false),
      render_pending(false), timer(0), hinstance(0), hwnd(0)
{
    // First of all: look for a previous instance of this program
    if (FindWindow(window_class_name, window_title)) {
        SetForegroundWindow(FindWindow(window_class_name, window_title));
        throw AppAlreadyExists();
    }
    
    // set self ptr
    if (self_ptr) throw AppError("app created twice");
    self_ptr = this;

    // Find HINSTANCE
    hinstance = GetModuleHandle(0);
    
    // Get screen size
    screen_width = GetSystemMetrics(SM_CXSCREEN);
    screen_height = GetSystemMetrics(SM_CYSCREEN);

    // Open up the GAPI library
    // (These have to be linked at run-time using LoadLibrary -- we can't just link to
    // gx.lib because of compiler incompatibilities.)
    HMODULE gapi_lib = LoadLibrary(TEXT("gx.dll"));
    MyImport(GAPI_GetDefaultKeys, gapi_lib, TEXT("?GXGetDefaultKeys@@YA?AUGXKeyList@@H@Z"));
    MyImport(GAPI_GetDisplayProperties, gapi_lib, TEXT("?GXGetDisplayProperties@@YA?AUGXDisplayProperties@@XZ"));
    MyImport(GAPI_OpenDisplay, gapi_lib, TEXT("?GXOpenDisplay@@YAHPAUHWND__@@K@Z"));
    MyImport(GAPI_OpenInput, gapi_lib, TEXT("?GXOpenInput@@YAHXZ"));
    MyImport(GAPI_CloseDisplay, gapi_lib, TEXT("?GXCloseDisplay@@YAHXZ"));
    MyImport(GAPI_CloseInput, gapi_lib, TEXT("?GXCloseInput@@YAHXZ"));
    MyImport(GAPI_BeginDraw, gapi_lib, TEXT("?GXBeginDraw@@YAPAXXZ"));
    MyImport(GAPI_EndDraw, gapi_lib, TEXT("?GXEndDraw@@YAHXZ"));
    MyImport(GAPI_Suspend, gapi_lib, TEXT("?GXSuspend@@YAHXZ"));
    MyImport(GAPI_Resume, gapi_lib, TEXT("?GXResume@@YAHXZ"));
    MyImport(GAPI_SetViewport, gapi_lib, TEXT("?GXSetViewport@@YAHKKKK@Z"));
    MyImport(GAPI_IsDisplayDRAMBuffer, gapi_lib, TEXT("?GXIsDisplayDRAMBuffer@@YAHXZ"));
    if (!GAPI_GetDefaultKeys || !GAPI_GetDisplayProperties || !GAPI_OpenDisplay
    || !GAPI_OpenInput || !GAPI_CloseDisplay || !GAPI_CloseInput
    || !GAPI_BeginDraw || !GAPI_EndDraw || !GAPI_Suspend || !GAPI_Resume 
    || !GAPI_SetViewport || !GAPI_IsDisplayDRAMBuffer) {
        throw AppError("failed to open GAPI (GX.DLL)");
    }

    // Ditto for AYGSHELL library
    HMODULE aygshell_lib = LoadLibrary(TEXT("aygshell.dll"));
    MyImport(SH_FullScreen, aygshell_lib, TEXT("SHFullScreen"));
    if (!SH_FullScreen) {
      throw AppError("failed to open AYGSHELL.DLL");
    }
}

//
// dtor
//

GapiApp2D::~GapiApp2D()
{
  // Shut everthing down
  if (display_open) {
    DestroyWindow(hwnd);
    GAPI_CloseInput();
    GAPI_CloseDisplay();
  }
  self_ptr = 0;
}


//
// openFullScreen
//

void GapiApp2D::openFullScreen(int, const char *)
{
    // Register a window class
    WNDCLASS wc = { 0L, GapiApp2D_WndProc, 0L, 0L, hinstance, 0, 0, 0, 0,
                    window_class_name };
    RegisterClass(&wc);

    // Create a window
    hwnd = CreateWindow(window_class_name,
                            window_title,
                WS_VISIBLE,
                CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
                             0, 0, hinstance, 0);

    // Open the GAPI display
    int success = GAPI_OpenDisplay(hwnd, GX_FULLSCREEN);
    if (!success) throw AppError("failed to open GAPI display");

    // Get the display properties
    // Only 16 bpp supported at present.
    props = GAPI_GetDisplayProperties();
    if (props.cBPP != 16) {
        GAPI_CloseDisplay();
        throw AppError("Sorry, this program only supports 16-bpp displays.");
    }

    if (props.ffFormat & kfDirect565) {
        rmask = 0xf800;
        rshift = 11;
        gmask = 0x07e0;
        gclamp = 2;
    } else if (props.ffFormat & kfDirect555) {
        rmask = 0x7c00;
        rshift = 10;
        gmask = 0x03e0;
        gclamp = 3;
    }

    // Capture user input
    GAPI_OpenInput();

    // Get a default key list
    keylist = GAPI_GetDefaultKeys(GX_NORMALKEYS);

    display_open = true;
}

//
// enterMainLoop -- windows message loop
//

void GapiApp2D::enterMainLoop()
{
    // Kickstart timers
    resetUpdateIntervals();
    
    // Enter main message loop
    MSG msg;
    msg.message = WM_PAINT;  // ensure msg.message is initialized to something.
    do {
        bool got_message;
        if (getCurrentUpdateInterval()==0 || wantRenderEachFrame() || render_pending) {
            got_message = PeekMessage(&msg, 0, 0, 0, PM_REMOVE);
        } else {
            // Use GetMessage if no updating or rendering is needed during idle time
            got_message = GetMessage(&msg, 0, 0, 0);
        }

        if (got_message) {
            // Check for WM_TIMER messages
            if (msg.message == WM_TIMER) {
                msg.hwnd = hwnd;
            }
            // Send message to window.
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            // No more messages waiting. Update and render a frame if wanted.
            if (getCurrentUpdateInterval() == 0) {
                update();
                if (wantRenderAfterUpdate()) render_pending = true;
            }
            if (wantRenderEachFrame() || render_pending) {
                render();
                render_pending = false;
            }
        }

        // Make sure we do renders in a timely fashion
        if (render_pending) {
            render(); 
            render_pending = false;
        }

    } while (msg.message != WM_QUIT);
}

void GapiApp2D::exitMainLoop()
{
    PostQuitMessage(0);
}


//
// wndProc
//

LRESULT GapiApp2D::wndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    const int MENU_HEIGHT = 26; 
    static RECT sm_rect, lrg_rect;
    
    switch (msg) {

    case WM_CREATE:
        GetWindowRect(hwnd, &sm_rect);
        GetWindowRect(hwnd, &lrg_rect);
        lrg_rect.top -= MENU_HEIGHT;
        lrg_rect.bottom += MENU_HEIGHT;
        break;
        
    case WM_CLOSE:
        DestroyWindow(hwnd);
        break;
        
    case WM_DESTROY:
        PostQuitMessage(0);
        break;

    case WM_PAINT:
        // Tell main loop to repaint the display (if wanted)
        if (wantRenderOnRepaint()) render_pending = true;
        break;
        
    case WM_ACTIVATE:
        {
            const int SHOWTASKBAR = 1;
            const int HIDETASKBAR = 2;
            const int SHOWSIPBUTTON = 4;
            const int HIDESIPBUTTON = 8;
            const int SHOWSTARTICON = 16;
            const int HIDESTARTICON = 32;
            if (LOWORD(wparam) == WA_INACTIVE) {
                if (SH_FullScreen) SH_FullScreen(hwnd, SHOWTASKBAR | SHOWSIPBUTTON |
                                                 SHOWSTARTICON);
                MoveWindow(hwnd, sm_rect.left, sm_rect.top, sm_rect.right,
                           sm_rect.bottom, true);
                if (GAPI_Suspend) GAPI_Suspend();
                deactivate();
                render_pending = false;
            } else {
                if (SH_FullScreen) SH_FullScreen(hwnd, HIDETASKBAR | HIDESIPBUTTON |
                                                 HIDESTARTICON);
                MoveWindow(hwnd, lrg_rect.left, lrg_rect.top, lrg_rect.right,
                               lrg_rect.bottom, true);
                if (GAPI_Resume) GAPI_Resume();             
                activate();
            }
        }
        break;

    // (Note -- we don't attempt to handle right-mouse-button messages in this App class)
    case WM_LBUTTONDOWN:
        mouseDown(LOWORD(lparam), HIWORD(lparam), false);
        break;

    case WM_LBUTTONUP:
        mouseUp(LOWORD(lparam), HIWORD(lparam), false);
        break;

    case WM_MOUSEMOVE:
        mouseMove(LOWORD(lparam), HIWORD(lparam));
        break;
        
    case WM_KEYDOWN:
        if (wparam == keylist.vkA) {
            key(Keycode('A'));
        } else if (wparam == keylist.vkB) {
            key(Keycode('B'));
        } else if (wparam == keylist.vkC) {
            key(Keycode('C'));
        } else if (wparam == keylist.vkStart) {
            key(Keycode('S'));
        } else if (wparam == keylist.vkUp) {
            key(KC_UP);
        } else if (wparam == keylist.vkDown) {
            key(KC_DOWN);
        } else if (wparam == keylist.vkLeft) {
            key(KC_LEFT);
        } else if (wparam == keylist.vkRight) {
            key(KC_RIGHT);
        }
        break;

    case WM_KEYUP:
        if (wparam == keylist.vkA) {
            keyUp(Keycode('A'));
        } else if (wparam == keylist.vkB) {
            keyUp(Keycode('B'));
        } else if (wparam == keylist.vkC) {
            keyUp(Keycode('C'));
        } else if (wparam == keylist.vkStart) {
            keyUp(Keycode('S'));
        } else if (wparam == keylist.vkUp) {
            keyUp(KC_UP);
        } else if (wparam == keylist.vkDown) {
            keyUp(KC_DOWN);
        } else if (wparam == keylist.vkLeft) {
            keyUp(KC_LEFT);
        } else if (wparam == keylist.vkRight) {
            keyUp(KC_RIGHT);
        }
        break;
        
    case WM_TIMER:
        // Timer event: it is time to call update()
        if (getCurrentUpdateInterval() > 0) { // (Check the timer is still valid)
            update();
            if (wantRenderAfterUpdate()) render_pending = true;
        }
        break;
    }

    return DefWindowProc(hwnd, msg, wparam, lparam);
}

//
// timer handling (resetUpdateIntervals)
//

void GapiApp2D::resetUpdateIntervals()
{
    if (timer) {
        KillTimer(0, timer);
    }

    if (getCurrentUpdateInterval() > 0) {
        timer = SetTimer(0, 0, getCurrentUpdateInterval(), 0);
    }
}

int GapiApp2D::getMilliseconds()
{
    return GetTickCount();
}

