/*
 * wce_serial.cpp
 *
 * (TODO) -- This code contains a lot of stuff copied from enet_stream.cpp (albeit with some
 * changes) -- should probably refactor into a common base class or something.
 *
 * Note -- we also have a dependency on "my_exceptions" but this could be taken out
 * if necessary
 * 
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "my_exceptions.hpp"
#include "wce_serial.hpp"

//
// Ctor, dtor (deal with opening and closing the COM port)
//

DWORD WceSerialBuf::threadRoutine(void *dat)
{
	WceSerialBuf *self = static_cast<WceSerialBuf*>(dat);

	// Open the port
	wchar_t filename[] = L"COM0:";
	filename[3] = L'0' + self->port;
	HANDLE handle = CreateFile(filename,
							   GENERIC_READ | GENERIC_WRITE,
							   0,
							   0,
							   OPEN_EXISTING,
							   0,
							   0);

	// Set comm timeouts
	COMMTIMEOUTS ct;
	memset(&ct, 0, sizeof(ct));
	// Set things up so that reads will return immediately, even if there were
	// no bytes available
	ct.ReadIntervalTimeout = MAXDWORD;
	ct.ReadTotalTimeoutConstant = 0;
	ct.ReadTotalTimeoutMultiplier = 0;
	// Writes will block for the following number of milliseconds
	// (don't set this to zero, otherwise writes will block forever, which could 
	// conceivably cause deadlocks if both sides were writing simultaneously)
	ct.WriteTotalTimeoutConstant = 20;   
	ct.WriteTotalTimeoutMultiplier = 0;  
	SetCommTimeouts(handle, &ct);

	// Write the port back to "self"
	EnterCriticalSection(&self->crit_sec);
	self->handle = handle;
	self->file_opened_yet = true;
	LeaveCriticalSection(&self->crit_sec);

	return 0;
}
	

WceSerialBuf::WceSerialBuf(int port_)
	: packet_state(P_WAITING), packet_idx(0), output_idx(0),
	  handle(0), connected(false), half_connected(false),
	  port(port_), thread(0), file_opened_yet(false)
{
	InitializeCriticalSection(&crit_sec);
	thread = CreateThread(0, 0, &threadRoutine, this, 0, 0);
}

WceSerialBuf::~WceSerialBuf()
{
	EnterCriticalSection(&crit_sec);
	bool ok = file_opened_yet;
	LeaveCriticalSection(&crit_sec);
	
	if (ok) {
		// Close the file handle
		CloseHandle(handle);
	} else {
		// Terminate the thread charged with opening the file handle
		// -- a bit drastic, but we have to do this, because otherwise
		// the thread will try to write into this object (which would
		// be bad since we have just been deleted!)

		// (TODO) I think this is slightly dodgy, eg it leaves an unclosable "CONSOLE"
		// window on my PocketPC. Should look into improvements sometime ...
		TerminateThread(thread, 0);
	}
	CloseHandle(thread);
	DeleteCriticalSection(&crit_sec);
}


//
// standard streambuf functions:
//

WceSerialBuf::int_type WceSerialBuf::overflow(WceSerialBuf::int_type meta)
{
	// This routine just extends the output buffer. The packet is only deemed to have
	// finished once "sync" is called. 
	
	if (meta == traits_type::eof()) {
		return meta;
	} else {
		char c = traits_type::to_char_type(meta);
		if (pptr() == epptr()) {
			if (output_buf.empty()) output_buf.resize(1);
			int old_size = output_buf.back().size()-2;
			if (old_size<0) old_size=0;
			const int new_size = old_size==0 ? 1024 : old_size*2;
			output_buf.back().resize(new_size+2);
			output_buf.back()[0] = 0;  // dummy area for the packet size
			output_buf.back()[1] = 0;  // (written when the packet is finished)
			setp( &(output_buf.back().front())+old_size+2, &(output_buf.back().back())+1 );
		}
		*(pptr()) = c;
		pbump(1);
		return traits_type::not_eof(meta);
	}
}

int WceSerialBuf::sync()
{
	// Finish off the current packet (if there is one):
	if (!output_buf.empty() && output_buf.back().size() > 2) {
		int sz = pptr() - (&output_buf.back().front());
		output_buf.back().resize(sz);
		sz -= 2;
		output_buf.back()[0] = (sz >> 8);
		output_buf.back()[1] = (sz & 0xff);
		output_buf.push_back(vector<char>());
		setp(0,0);
	}
	
	// Read/write data to/from the serial port
	if (!writeOutputBuffer()) return -1;
	if (!fillInputBuffer()) return -1;

	return 0;
}

WceSerialBuf::int_type WceSerialBuf::underflow()
{
	// First read data from serial port
	if (!fillInputBuffer()) return traits_type::eof();

	// Now return a value from the buffer.
	if (gptr() == egptr()) return traits_type::eof();
	return traits_type::to_int_type(*gptr());
}

streamsize WceSerialBuf::showmanyc()
{
	// First read data from serial port
	if (!fillInputBuffer()) return 0;

	// Now return how many characters are left on the current packet
	return (egptr() - gptr());
}


//
// tryConnect -- handshaking sequence
//

bool WceSerialBuf::tryConnect()
{
	// Check if our "opener" thread is done. If not, return false immediately.
	EnterCriticalSection(&crit_sec);
	bool ok = file_opened_yet;
	LeaveCriticalSection(&crit_sec);
	if (!ok) return false;

	// check if we are already connected
	if (connected) return true;

	// Handshaking protocol. This is crude but works for now.
	
	// Receive character from other side
	char c = 0;
	DWORD bytes_read = 0;
	bool success = ReadFile(handle, &c, 1, &bytes_read, 0);
	if (success && bytes_read == 1) {

		if (c == 'S' || c == 'C') {

			// There's life on the other side of the connection...

			// Recall: "half connected" means we have sent our S
			// "connected" means we have both received an S and sent one too
			
			// Send out our 'S' if we haven't already
			if (!half_connected) {
				DWORD bytes_written = 0;
				char s = 'S';
				success = WriteFile(handle, &s, 1, &bytes_written, 0);
				if (!success || bytes_written != 1)
					throw UnexpectedError("handshaking failed (error 1)");
				half_connected = true;
			}

			// If an 'S' was received then we can go into full connected mode.
			if (c=='S') {
				connected = true;
				return true;
			}

			// If a 'C' was received -- then we are only half connected. We must wait for
			// the other side to send an 'S'. We know that they will, because (a) we have
			// sent out our 'S' and (b) we know that there's life on the other side so they
			// will definitely have received the 'S' we just sent....
			
		} else {
			// Hmm, unexpected character received.
			throw UnexpectedError("handshaking failed (error 2)");
		}				
	}

	if (!connected && !half_connected) {
		// We haven't even received any 'C' from the other side yet.
		// Try to send out a 'C', see if it gets any response ...
		DWORD bytes_written = 0;
		c = 'C';
		success = WriteFile(handle, &c, 1, &bytes_written, 0);
		if (!success /* || bytes_written != 1 */) {
			throw UnexpectedError("handshaking failed (error 3)");
		}
	}
	
	return false;
}


//
// fillInputBuffer
// Read data from the COM port
// Returns true if successful or false on error
//

bool WceSerialBuf::fillInputBuffer()
{
	if (packet_state == P_WAITING) {
		// Read first half of the packet size
		unsigned char c = 0;
		DWORD nbytes = 0;
		bool success = ReadFile(handle, &c, 1, &nbytes, 0);
		//		if (!success) return false; // Read error on COM port
		if (!success) throw UnexpectedError("COM error 1");
		if (nbytes != 0) {
			packet_state = P_HALFWAY;
			packet_idx = c;
		}
	}

	if (packet_state == P_HALFWAY) {
		// Read second half of the packet size
		unsigned char c = 0;
		DWORD nbytes = 0;
		bool success = ReadFile(handle, &c, 1, &nbytes, 0);
		//		if (!success) return false;
		if (!success) throw UnexpectedError("COM error 2");
		if (nbytes != 0) {
			packet_state = P_RECEIVING;
			input_buf.push_back(vector<char>());
			input_buf.back().resize((packet_idx<<8) + c);
			ASSERT(!input_buf.back().empty());
			packet_idx = 0;
		}
	}

	if (packet_state == P_RECEIVING) {
		DWORD nbytes = 0;
		bool success = ReadFile(handle, &input_buf.back()[packet_idx],
								input_buf.back().size() - packet_idx,
								&nbytes, 0);
		//		if (!success) return false;
		if (!success) throw UnexpectedError("COM error 3");
		packet_idx += nbytes;
		ASSERT(packet_idx <= input_buf.back().size());
		if (packet_idx == input_buf.back().size()) {
			// Done reading packet!
			packet_state = P_WAITING;
		}
	}

	// We now have to sort out the get pointers:
	if (gptr() == egptr()) {
		if (!input_buf.empty() && gptr() >= &input_buf.front().front()
		&& gptr() <= (&input_buf.front().back())+1) {
			// Get pointer is within the first packet (presumably at the end of it)
			// so we must delete the first packet (we have read all of it).
			input_buf.pop_front();
		}

		if (input_buf.empty() || input_buf.size()==1 && packet_state != P_WAITING) {
			// We have no packets, OR only one packet which is the currently-being-read
			// packet. Clear the get pointers.
			setg(0,0,0);
		} else {
			// Set the get pointers to point into the front packet.
			setg(&input_buf.front().front(), &input_buf.front().front(),
				 (&input_buf.front().back())+1);
		}
	}
	
	return true;
}

//
// writeOutputBuffer
// Flushes output buffers to the COM port, as far as possible.
// Returns true if successful or false if there was a write error.
//

bool WceSerialBuf::writeOutputBuffer()
{
	while (output_buf.size() > 1) {
		DWORD nbytes = 0;
		bool success = WriteFile(handle, &output_buf.front()[output_idx],
								 output_buf.front().size() - output_idx,
								 &nbytes, 0);
		//		if (!success) return false;
		if (!success) throw UnexpectedError("COM error 11");
		output_idx += nbytes;
		ASSERT(output_idx <= output_buf.front().size());
		if (output_idx == output_buf.front().size()) {
			output_buf.pop_front();
			output_idx = 0;
			continue;
		} else {
			break;
		}
	}
	return true;
}
